// copyright 2022 zamfofex
// license: AGPLv3 or later

#include <mimimi/chapters.h>
#include <mimimi/senran-kagura/characters.h>
#include <mimimi/appearances.h>
#include <mimimi/allocators.h>
#include <mimimi/ground.h>
#include <mimimi/compound-behaviors.h>
#include <mimimi/geometry.h>
#include <mimimi/assets.h>
#include <mimimi/engines.h>
#include <mimimi/controls.h>
#include <mimimi/sprites.h>
#include <mimimi/cameras.h>
#include <mimimi/behaviors.h>
#include <mimimi/font.h>
#include <mimimi/fonts.h>
#include <mimimi/animations.h>
#include <mimimi/collisions.h>
#include <mimimi/test.h>

#include "ground.c"

static void mimimi_test_animation_set(struct mimimi_animation_set *set, struct mimimi_behavior *compound, struct mimimi_allocator *allocator)
{
	int animation_count = 8;
	int image_count = 6;
	int width = 48;
	int height = 48;
	
	int total_animation_count = 2 + animation_count;
	int total_image_count = 2 + image_count * animation_count + image_count;
	int total_color_count = width * height * total_image_count;
	
	struct mimimi_animation *animations = mimimi_compound_allocate(compound, total_animation_count * sizeof *animations, allocator);
	struct mimimi_image *images = mimimi_compound_allocate(compound, total_image_count * sizeof *images, allocator);
	unsigned char *colors = mimimi_compound_allocate(compound, total_color_count, allocator);
	
	for (int i = 0 ; i < total_color_count ; i++)
		colors[i] = 0;
	
	set->standing = animations;
	set->standing_animation_count = animation_count;
	animations += animation_count;
	
	for (int i = 0 ; i < animation_count ; i++)
	{
		struct mimimi_animation *animation = set->standing + i;
		animation->count = image_count;
		animation->images = images;
		images += image_count;
		
		for (int i = 0 ; i < image_count ; i++)
		{
			struct mimimi_image *image = animation->images + i;
			image->width = width;
			image->height = height;
			image->colors = colors;
			colors += width * height;
		}
	}
	
	struct mimimi_animation *knocked = animations++;
	knocked->count = 1;
	knocked->images = images++;
	knocked->images[0].width = width;
	knocked->images[0].height = height;
	knocked->images[0].colors = colors;
	colors += width * height;
	
	struct mimimi_animation *falling = animations++;
	falling->count = 1;
	falling->images = images++;
	falling->images[0].width = width;
	falling->images[0].height = height;
	falling->images[0].colors = colors;
	colors += width * height;
	
	set->knocked = knocked;
	set->falling = falling;
	
	for (int i = 0 ; i < width * height ; i++)
	{
		knocked->images[0].colors[i] = 0;
		falling->images[0].colors[i] = 0;
	}
	
	set->jumping = images;
	set->jumping_image_count = image_count;
	images += image_count;
	
	for (int i = 0 ; i < image_count ; i++)
	{
		struct mimimi_image *image = set->jumping + i;
		image->width = width;
		image->height = height;
		image->colors = colors;
		colors += width * height;
	}
}

static struct mimimi_appearance *mimimi_test_appearance(struct mimimi_behavior *compound, struct mimimi_allocator *allocator)
{
	struct mimimi_appearance *appearance = mimimi_compound_allocate(compound, sizeof *appearance, allocator);
	
	mimimi_test_animation_set(&appearance->left, compound, allocator);
	mimimi_test_animation_set(&appearance->right, compound, allocator);
	
	return appearance;
}

struct mimimi_test_sprite
{
	int x, y;
	struct mimimi_model *model;
};

static void mimimi_test_player(struct mimimi_behavior *compound, struct mimimi_chapter *chapter, struct mimimi_sprite *player, struct mimimi_allocator *allocator)
{
	struct mimimi_history *history = mimimi_compound_allocate(compound, sizeof *history, allocator);
	mimimi_compound(compound, mimimi_history(history, chapter, allocator));
	
	struct mimimi_jump *jump = mimimi_compound_allocate(compound, sizeof *jump, allocator);
	struct mimimi_behavior *jump_behavior = mimimi_jump(jump, player->physics, allocator);
	mimimi_compound(compound, mimimi_finalizer(jump_behavior->data, jump_behavior->finish, allocator));
	
	// struct mimimi_dash *dash = mimimi_compound_allocate(compound, sizeof *dash, allocator);
	// struct mimimi_behavior *dash_behavior = mimimi_dash(dash, player->physics, player->walk, allocator);
	// mimimi_compound(compound, mimimi_finalizer(dash_behavior->data, dash_behavior->finish, allocator));
	// struct mimimi_behavior *throttled_dash = mimimi_throttle(compound, dash_behavior, 32, allocator);
	
	struct mimimi_behavior *jump_choice = mimimi_choice(&player->physics->airborne, mimimi_noop, jump_behavior, allocator);
	
	struct mimimi_controls *mirai_controls = mimimi_compound_allocate(compound, sizeof *mirai_controls, allocator);
	player->awake_behavior = mimimi_controls(mirai_controls, player->walk, history, allocator);
	mimimi_compound(compound, mimimi_finalizer(player->awake_behavior->data, player->awake_behavior->finish, allocator));
	mirai_controls->jump = jump_choice;
}

struct mimimi_chapter *mimimi_test(struct mimimi_allocator *allocator)
{
	struct mimimi_behavior *compound = mimimi_compound_behavior(allocator);
	
	struct mimimi_engine *engine = mimimi_compound_allocate(compound, sizeof *engine, allocator);
	
	struct mimimi_chapter *chapter = mimimi_compound_allocate(compound, sizeof *chapter, allocator);
	chapter->behavior = compound;
	chapter->engine = engine;
	
	struct mimimi_position *camera = mimimi_compound_allocate(compound, sizeof *camera, allocator);
	
	struct mimimi_test_sprite infos[] =
	{
		{11776, 7040, mimimi_mirai_flash},
		// {10000, 6912, mimimi_ryoubi_flash},
		{11000, 7168, mimimi_ryouna_flash},
	};
	
	int count = sizeof infos / sizeof *infos;
	
	struct mimimi_position **positions = mimimi_compound_allocate(compound, count * sizeof *positions, allocator);
	mimimi_compound(compound, mimimi_collision(count, positions, 128, 256, 512, 64, allocator));
	
	int *directions = mimimi_compound_allocate(compound, count * sizeof *directions, allocator);
	
	struct mimimi_sprite **sprites = (allocator->allocate)(count * sizeof *sprites);
	
	for (int i = 0 ; i < count ; i++)
	{
		struct mimimi_test_sprite *info = infos + i;
		
		struct mimimi_sprite *sprite = mimimi_sprite(&mimimi_test_ground, info->x, info->y, 80, 250, allocator);
		mimimi_compound(compound, sprite->behavior);
		sprites[i] = sprite;
		
		if (i == 0)
		{
			mimimi_test_player(compound, chapter, sprite, allocator);
			mimimi_compound(compound, mimimi_camera(camera, sprite->position, 0, -512, allocator));
		}
		else
		{
			mimimi_compound(compound, mimimi_stationary(directions + i, &sprite->position->x, &sprites[i]->position->x, allocator));
		}
		
		positions[i] = sprite->position;
	}
	
	struct mimimi_image *background = mimimi_compound_allocate(compound, sizeof *background, allocator);
	background->colors = mimimi_compound_allocate(compound, mimimi_test_ground.width * 16 * mimimi_test_ground.height * 16, allocator);
	
	background->width = mimimi_test_ground.width * 16;
	background->height = mimimi_test_ground.height * 16;
	
	static unsigned char ground_colors[] = {0xFA, 0xEF, 0xDB, 0xEA, 0xE4, 0xBC, 0xEB};
	
	for (int y = 0 ; y < mimimi_test_ground.height * 16 ; y++)
	for (int x = 0 ; x < mimimi_test_ground.width * 16 ; x++)
	{
		unsigned char color = ground_colors[mimimi_test_ground.tiles[x / 16 + y / 16 * mimimi_test_ground.width]];
		background->colors[x + y * mimimi_test_ground.width * 16] = color;
	}
	
	mimimi_compound(compound, mimimi_background(engine, camera, background, 0, allocator));
	
	for (int i = 0 ; i < count ; i++)
	{
		struct mimimi_test_sprite *info = infos + i;
		struct mimimi_sprite *sprite = sprites[i];
		
		struct mimimi_appearance *appearance = mimimi_test_appearance(compound, allocator);
		mimimi_appearance(appearance, info->model, 24, 48, allocator);
		mimimi_compound(compound, mimimi_display(engine, camera, sprite, appearance, directions + i, allocator));
	}
	
	(allocator->deallocate)(sprites);
	
	struct mimimi_image *overlay = mimimi_compound_allocate(compound, sizeof *overlay, allocator);
	overlay->width = 256;
	overlay->height = 64;
	overlay->colors = mimimi_compound_allocate(compound, overlay->width * overlay->height, allocator);
	
	for (int i = 0 ; i < overlay->width * overlay->height ; i++)
		overlay->colors[i] = 0;
	
	struct mimimi_image *glyphs = (*allocator->allocate)(mimimi_font_image_count * sizeof *glyphs);
	unsigned char *colors = (*allocator->allocate)(mimimi_font_color_count);
	
	for (int i = 0 ; i < mimimi_font_image_count ; i++)
		glyphs[i].colors = colors + i * mimimi_font_image_color_count;
	
	mimimi_glyphs(glyphs, mimimi_font, 0xDF);
	
	mimimi_text(overlay, glyphs, 16, 16, "Mirai's Miscellaneous Misadventures M26");
	mimimi_text(overlay, glyphs, 16, 32, "a Senran Kagura fangame, June 12 of 2022");
	
	(*allocator->deallocate)(colors);
	(*allocator->deallocate)(glyphs);
	
	mimimi_compound(compound, mimimi_overlay(engine, overlay, 0, 0, allocator));
	
	return chapter;
}
